@@ -26,10 +26,89 @@
  */
 #include "intel_xvmc.h"
 
+/* global */
 struct _intel_xvmc_driver *xvmc_driver = NULL;
+
+/* Lookup tables to speed common calculations for coded_block_pattern */
+/* each block is ((8*8) * sizeof(short)) */
+unsigned int mb_bytes_420[] = {
+    0, /* 0 */
+    128, /* 1 */
+    128, /* 10 */
+    256, /* 11 */
+    128, /* 100 */
+    256, /* 101 */
+    256, /* 110 */
+    384, /* 111 */
+    128, /* 1000 */
+    256, /* 1001 */
+    256, /* 1010 */
+    384, /* 1011 */
+    256, /* 1100 */
+    384, /* 1101 */
+    384, /* 1110 */
+    512, /* 1111 */
+    128, /* 10000 */
+    256, /* 10001 */
+    256, /* 10010 */
+    384, /* 10011 */
+    256, /* 10100 */
+    384, /* 10101 */
+    384, /* 10110 */
+    512, /* 10111 */
+    256, /* 11000 */
+    384, /* 11001 */
+    384, /* 11010 */
+    512, /* 11011 */
+    384, /* 11100 */
+    512, /* 11101 */
+    512, /* 11110 */
+    640, /* 11111 */
+    128, /* 100000 */
+    256, /* 100001 */
+    256, /* 100010 */
+    384, /* 100011 */
+    256, /* 100100 */
+    384, /* 100101 */
+    384, /* 100110 */
+    512, /* 100111 */
+    256, /* 101000 */
+    384, /* 101001 */
+    384, /* 101010 */
+    512, /* 101011 */
+    384, /* 101100 */
+    512, /* 101101 */
+    512, /* 101110 */
+    640, /* 101111 */
+    256, /* 110000 */
+    384, /* 110001 */
+    384, /* 110010 */
+    512, /* 110011 */
+    384, /* 110100 */
+    512, /* 110101 */
+    512, /* 110110 */
+    640, /* 110111 */
+    384, /* 111000 */
+    512, /* 111001 */
+    512, /* 111010 */
+    640, /* 111011 */
+    512, /* 111100 */
+    640, /* 111101 */
+    640, /* 111110 */
+    768  /* 111111 */
+};
+
+int DEBUG;
+
 static int error_base;
 static int event_base;
 
+static void intel_xvmc_debug_init(void)
+{
+    if (getenv("INTEL_XVMC_DEBUG"))
+	DEBUG = 1;
+}
+
 /* locking */
 static void intel_xvmc_try_heavy_lock(drm_context_t ctx)
 {
@@ -185,7 +264,7 @@ intel_xvmc_surface_ptr intel_xvmc_find_surface(XID id)
 *        returned by XvMCListSurfaceTypes.
 * Returns: Status
 */
-Status XvMCCreateContext(Display *display, XvPortID port,
+_X_EXPORT Status XvMCCreateContext(Display *display, XvPortID port,
                          int surface_type_id, int width, int height,
                          int flags, XvMCContext *context)
 {
@@ -211,6 +290,8 @@ Status XvMCCreateContext(Display *display, XvPortID port,
         return BadValue;
     }
 
+    intel_xvmc_debug_init();
+
     /* Open DRI Device */
     if((fd = drmOpen("i915", NULL)) < 0) {
         XVMC_ERR("DRM Device could not be opened.");
@@ -260,6 +341,8 @@ Status XvMCCreateContext(Display *display, XvPortID port,
 		xvmc_driver = &i915_xvmc_mc_driver;
 		break;
 	    case XVMC_I965_MPEG2_MC:
+		xvmc_driver = &i965_xvmc_mc_driver;
+		break;
 	    case XVMC_I945_MPEG2_VLD:
 	    case XVMC_I965_MPEG2_VLD:
 	    default:
@@ -360,6 +443,8 @@ Status XvMCCreateContext(Display *display, XvPortID port,
 
     intelInitBatchBuffer();
 
+    intel_xvmc_dump_open();
+
     return Success;
 }
 
@@ -372,14 +457,14 @@ Status XvMCCreateContext(Display *display, XvPortID port,
  *   context - The context to be destroyed.
  *
  */
-Status XvMCDestroyContext(Display *display, XvMCContext *context)
+_X_EXPORT Status XvMCDestroyContext(Display *display, XvMCContext *context)
 {
     Status ret;
-    int screen = DefaultScreen(display);
+    int screen;
 
     if (!display || !context)
         return XvMCBadContext;
-
+    screen = DefaultScreen(display);
     ret = (xvmc_driver->destroy_context)(display, context);
     if (ret) {
 	XVMC_ERR("destroy context fail\n");
@@ -407,6 +492,8 @@ Status XvMCDestroyContext(Display *display, XvMCContext *context)
 	xvmc_driver->fd = -1;
 
 	intelFiniBatchBuffer();
+
+	intel_xvmc_dump_close();
     }
     return Success;
 }
@@ -414,7 +501,7 @@ Status XvMCDestroyContext(Display *display, XvMCContext *context)
 /*
  * Function: XvMCCreateSurface
  */
-Status XvMCCreateSurface(Display *display, XvMCContext *context, XvMCSurface *surface)
+_X_EXPORT Status XvMCCreateSurface(Display *display, XvMCContext *context, XvMCSurface *surface)
 {
     Status ret;
     int priv_count;
@@ -463,7 +550,7 @@ Status XvMCCreateSurface(Display *display, XvMCContext *context, XvMCSurface *su
 /*
  * Function: XvMCDestroySurface
  */
-Status XvMCDestroySurface(Display *display, XvMCSurface *surface)
+_X_EXPORT Status XvMCDestroySurface(Display *display, XvMCSurface *surface)
 {
     intel_xvmc_surface_ptr intel_surf;
 
@@ -489,16 +576,17 @@ Status XvMCDestroySurface(Display *display, XvMCSurface *surface)
 /*
  * Function: XvMCCreateBlocks
  */
-Status XvMCCreateBlocks(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCCreateBlocks(Display *display, XvMCContext *context,
                         unsigned int num_blocks,
                         XvMCBlockArray *block)
 {
+    Status ret;
     if (!display || !context || !num_blocks || !block)
         return BadValue;
 
     memset(block, 0, sizeof(XvMCBlockArray));
 
-    if (!(block->blocks = (short *)malloc(num_blocks << 6 * sizeof(short))))
+    if (!(block->blocks = (short *)malloc((num_blocks << 6) * sizeof(short))))
         return BadAlloc;
 
     block->num_blocks = num_blocks;
@@ -511,9 +599,10 @@ Status XvMCCreateBlocks(Display *display, XvMCContext *context,
 /*
  * Function: XvMCDestroyBlocks
  */
-Status XvMCDestroyBlocks(Display *display, XvMCBlockArray *block)
+_X_EXPORT Status XvMCDestroyBlocks(Display *display, XvMCBlockArray *block)
 {
-    if (!display || block)
+    Status ret;
+    if (!display || !block)
         return BadValue;
 
     if (block->blocks)
@@ -530,7 +619,7 @@ Status XvMCDestroyBlocks(Display *display, XvMCBlockArray *block)
 /*
  * Function: XvMCCreateMacroBlocks
  */
-Status XvMCCreateMacroBlocks(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCCreateMacroBlocks(Display *display, XvMCContext *context,
                              unsigned int num_blocks,
                              XvMCMacroBlockArray *blocks)
 {
@@ -553,7 +642,7 @@ Status XvMCCreateMacroBlocks(Display *display, XvMCContext *context,
 /*
  * Function: XvMCDestroyMacroBlocks
  */
-Status XvMCDestroyMacroBlocks(Display *display, XvMCMacroBlockArray *block)
+_X_EXPORT Status XvMCDestroyMacroBlocks(Display *display, XvMCMacroBlockArray *block)
 {
     if (!display || !block)
         return BadValue;
@@ -575,7 +664,7 @@ Status XvMCDestroyMacroBlocks(Display *display, XvMCMacroBlockArray *block)
  *  macroblock structures it dispatched the hardware commands to execute
  *  them.
  */
-Status XvMCRenderSurface(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCRenderSurface(Display *display, XvMCContext *context,
                          unsigned int picture_structure,
                          XvMCSurface *target_surface,
                          XvMCSurface *past_surface,
@@ -595,6 +684,10 @@ Status XvMCRenderSurface(Display *display, XvMCContext *context,
     if (!target_surface)
 	return XvMCBadSurface;
 
+    intel_xvmc_dump_render(context, picture_structure, target_surface,
+	    past_surface, future_surface, flags, num_macroblocks,
+	    first_macroblock, macroblock_array, blocks);
+
     ret = (xvmc_driver->render_surface)(display, context, picture_structure,
 	    target_surface, past_surface, future_surface, flags,
 	    num_macroblocks, first_macroblock, macroblock_array,
@@ -632,7 +725,7 @@ Status XvMCRenderSurface(Display *display, XvMCContext *context,
  *	XVMC_BOTTOM_FIELD - Display only the Bottom Field of the surface.
  *	XVMC_FRAME_PICTURE - Display both fields or frame.
  */
-Status XvMCPutSurface(Display *display,XvMCSurface *surface,
+_X_EXPORT Status XvMCPutSurface(Display *display,XvMCSurface *surface,
                       Drawable draw, short srcx, short srcy,
                       unsigned short srcw, unsigned short srch,
                       short destx, short desty,
@@ -661,19 +754,23 @@ Status XvMCPutSurface(Display *display,XvMCSurface *surface,
 	intel_surf->gc = XCreateGC(display, draw, 0, NULL);
     }
     intel_surf->last_draw = draw;
-
     /* fill intel_surf->data */
+    if (0)
+    {
+	drmVBlank vbl;
+	vbl.request.type = DRM_VBLANK_RELATIVE;
+	vbl.request.sequence = 1;
+	drmWaitVBlank(xvmc_driver->fd, &vbl);
+    }
     ret = (xvmc_driver->put_surface)(display, surface, draw, srcx, srcy,
 	    srcw, srch, destx, desty, destw, desth, flags, &intel_surf->data);
     if (ret) {
 	XVMC_ERR("put surface fail\n");
 	return ret;
     }
-
     ret = XvPutImage(display, context->port, draw, intel_surf->gc,
 	    intel_surf->image, srcx, srcy, srcw, srch, destx, desty,
 	    destw, desth);
-
     return ret;
 }
 
@@ -683,7 +780,7 @@ Status XvMCPutSurface(Display *display,XvMCSurface *surface,
  *   display - Connection to the X server
  *   surface - The surface to synchronize
  */
-Status XvMCSyncSurface(Display *display, XvMCSurface *surface)
+_X_EXPORT Status XvMCSyncSurface(Display *display, XvMCSurface *surface)
 {
     Status ret;
     int stat = 0;
@@ -708,7 +805,7 @@ Status XvMCSyncSurface(Display *display, XvMCSurface *surface)
  *   surface - Surface to flush
  * Returns: Status
  */
-Status XvMCFlushSurface(Display * display, XvMCSurface *surface)
+_X_EXPORT Status XvMCFlushSurface(Display * display, XvMCSurface *surface)
 {
     if (!display || !surface)
 	return XvMCBadSurface;
@@ -727,7 +824,7 @@ Status XvMCFlushSurface(Display * display, XvMCSurface *surface)
  *    XVMC_DISPLAYING - The surface is currently being displayed or a
  *                     display is pending.
  */
-Status XvMCGetSurfaceStatus(Display *display, XvMCSurface *surface, int *stat)
+_X_EXPORT Status XvMCGetSurfaceStatus(Display *display, XvMCSurface *surface, int *stat)
 {
     Status ret;
 
@@ -752,7 +849,7 @@ Status XvMCGetSurfaceStatus(Display *display, XvMCSurface *surface, int *stat)
  *
  * Returns: Status
  */
-Status XvMCHideSurface(Display *display, XvMCSurface *surface)
+_X_EXPORT Status XvMCHideSurface(Display *display, XvMCSurface *surface)
 {
     int stat = 0;
     Status ret;
@@ -791,7 +888,7 @@ Status XvMCHideSurface(Display *display, XvMCSurface *surface)
  *
  * Returns: Status
  */
-Status XvMCCreateSubpicture(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCCreateSubpicture(Display *display, XvMCContext *context,
                             XvMCSubpicture *subpicture,
                             unsigned short width, unsigned short height,
                             int xvimage_id)
@@ -812,7 +909,7 @@ Status XvMCCreateSubpicture(Display *display, XvMCContext *context,
  *
  * Returns: Status
  */
-Status XvMCClearSubpicture(Display *display, XvMCSubpicture *subpicture,
+_X_EXPORT Status XvMCClearSubpicture(Display *display, XvMCSubpicture *subpicture,
                            short x, short y,
                            unsigned short width, unsigned short height,
                            unsigned int color)
@@ -836,7 +933,7 @@ Status XvMCClearSubpicture(Display *display, XvMCSubpicture *subpicture,
  *
  * Returns: Status
  */
-Status XvMCCompositeSubpicture(Display *display, XvMCSubpicture *subpicture,
+_X_EXPORT Status XvMCCompositeSubpicture(Display *display, XvMCSubpicture *subpicture,
                                XvImage *image,
                                short srcx, short srcy,
                                unsigned short width, unsigned short height,
@@ -856,7 +953,7 @@ Status XvMCCompositeSubpicture(Display *display, XvMCSubpicture *subpicture,
  *
  * Returns: Status
  */
-Status XvMCDestroySubpicture(Display *display, XvMCSubpicture *subpicture)
+_X_EXPORT Status XvMCDestroySubpicture(Display *display, XvMCSubpicture *subpicture)
 {
     XVMC_ERR("XvMCDestroySubpicture not implemented!");
     return BadValue;
@@ -873,7 +970,7 @@ Status XvMCDestroySubpicture(Display *display, XvMCSubpicture *subpicture)
  *     is num_palette_entries * entry_bytes in size.
  * Returns: Status
  */
-Status XvMCSetSubpicturePalette(Display *display, XvMCSubpicture *subpicture,
+_X_EXPORT Status XvMCSetSubpicturePalette(Display *display, XvMCSubpicture *subpicture,
                                 unsigned char *palette)
 {
     XVMC_ERR("XvMCSetSubpicturePalette not implemented!");
@@ -903,7 +1000,7 @@ Status XvMCSetSubpicturePalette(Display *display, XvMCSubpicture *subpicture,
  *
  * Returns: Status
  */
-Status XvMCBlendSubpicture(Display *display, XvMCSurface *target_surface,
+_X_EXPORT Status XvMCBlendSubpicture(Display *display, XvMCSurface *target_surface,
                            XvMCSubpicture *subpicture,
                            short subx, short suby,
                            unsigned short subw, unsigned short subh,
@@ -940,7 +1037,7 @@ Status XvMCBlendSubpicture(Display *display, XvMCSurface *target_surface,
  *
  * Returns: Status
  */
-Status XvMCBlendSubpicture2(Display *display,
+_X_EXPORT Status XvMCBlendSubpicture2(Display *display,
                             XvMCSurface *source_surface,
                             XvMCSurface *target_surface,
                             XvMCSubpicture *subpicture,
@@ -963,7 +1060,7 @@ Status XvMCBlendSubpicture2(Display *display,
  *
  * Returns: Status
  */
-Status XvMCSyncSubpicture(Display *display, XvMCSubpicture *subpicture)
+_X_EXPORT Status XvMCSyncSubpicture(Display *display, XvMCSubpicture *subpicture)
 {
     XVMC_ERR("XvMCSyncSubpicture not implemented!");
     return BadValue;
@@ -980,7 +1077,7 @@ Status XvMCSyncSubpicture(Display *display, XvMCSubpicture *subpicture)
  *
  * Returns: Status
  */
-Status XvMCFlushSubpicture(Display *display, XvMCSubpicture *subpicture)
+_X_EXPORT Status XvMCFlushSubpicture(Display *display, XvMCSubpicture *subpicture)
 {
     XVMC_ERR("XvMCFlushSubpicture not implemented!");
     return BadValue;
@@ -1000,7 +1097,7 @@ Status XvMCFlushSubpicture(Display *display, XvMCSubpicture *subpicture)
  *
  * Returns: Status
  */
-Status XvMCGetSubpictureStatus(Display *display, XvMCSubpicture *subpicture,
+_X_EXPORT Status XvMCGetSubpictureStatus(Display *display, XvMCSubpicture *subpicture,
                                int *stat)
 {
     XVMC_ERR("XvMCGetSubpictureStatus not implemented!");
@@ -1021,7 +1118,7 @@ Status XvMCGetSubpictureStatus(Display *display, XvMCSubpicture *subpicture,
  * Returns:
  *  An array of XvAttributes.
  */
-XvAttribute *XvMCQueryAttributes(Display *display, XvMCContext *context,
+_X_EXPORT XvAttribute *XvMCQueryAttributes(Display *display, XvMCContext *context,
                                  int *number)
 {
     /* now XvMC has no extra attribs than Xv */
@@ -1042,7 +1139,7 @@ XvAttribute *XvMCQueryAttributes(Display *display, XvMCContext *context,
  * Returns:
  *  Status
  */
-Status XvMCSetAttribute(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCSetAttribute(Display *display, XvMCContext *context,
                         Atom attribute, int value)
 {
     return Success;
@@ -1062,7 +1159,7 @@ Status XvMCSetAttribute(Display *display, XvMCContext *context,
  * Returns:
  *  Status
  */
-Status XvMCGetAttribute(Display *display, XvMCContext *context,
+_X_EXPORT Status XvMCGetAttribute(Display *display, XvMCContext *context,
                         Atom attribute, int *value)
 {
     return Success;
